@@ -26,10 +26,74 @@
  * EMail: roseg@apsis.ch
  */
 
-static char *rcs_id = "$Id: http.c,v 1.6 2003/11/30 22:56:26 roseg Rel $";
+static char *rcs_id = "$Id: http.c,v 1.10 2006/02/01 11:19:53 roseg Rel $";
 
 /*
  * $Log: http.c,v $
+ * Revision 1.10  2006/02/01 11:19:53  roseg
+ * Enhancements:
+ *   added NoDaemon configuration directive (replaces compile-time switch)
+ *   added LogFacility configuration directive (replaces compile-time switch)
+ *   added user name logging
+ *
+ * Bug fixes:
+ *   fixed problem with the poll() code
+ *   fixed problem with empty list in gethostbyname()
+ *   added call to setsid() if daemon
+ *   conflicting headers are removed (Content-length - Transfer-encoding)
+ *
+ * Last release in the 1.x series.
+ *
+ * Revision 1.9  2005/06/01 15:01:53  roseg
+ * Enhancements:
+ *   Added the VerifyList configuration flag (CA root certs + CRL)
+ *   CRL checking code
+ *   RewriteRedirect 2 - ignores port value for host matching
+ *   Added -c flag (check-only mode)
+ *   Added -v flag (verbose mode)
+ *   Added -p flag for pid file name
+ *
+ * Bug fixes:
+ *   fixed a potential buffer overflow problem (in checking the Host header)
+ *   added call to SSL_library_init
+ *   added a check for MSIE before forcing SSL shutdown
+ *   X-SSL-Cipher header is added only if HTTPSHeaders is non-zero
+ *   added code for shorter linger on badly closed connections (IE work-around)
+ *   fixed the locking for session checking (mutex_lock/unlock)
+ *
+ * Revision 1.8  2004/11/04 13:37:07  roseg
+ * Changes:
+ * - added support for non-blocking connect(2)
+ * - added support for 414 - Request URI too long
+ * - added RedirectRewrite directive - to prevent redirect changes
+ * - added support for NoHTTPS11 value 2 (for MSIE clients only)
+ * - added support for HTTPSHeaders 3 (no verify)
+ *
+ * Problems fixed:
+ * - fixed bug if multiple listening ports/addresses
+ * - fixed memory leak in SSL
+ * - flush stdout (if used) after each log message
+ * - assumes only 304, 305 and 306 codes to have no content
+ * - fixed problem with delays in 302 without content
+ * - fixed problem with time-outs in HTTPS
+ *
+ * Enhancements:
+ * - improved threads detection code in autoconf
+ * - added supervisor process disable configuration flag
+ * - tweak for the Location rewriting code (only look at current GROUP)
+ * - improved print-out for client certificate information
+ *
+ * Revision 1.7  2004/03/24 06:59:59  roseg
+ * Fixed bug in X-SSL-CIPHER description
+ * Changed README to stx format for consistency
+ * Addedd X-SSL-certificate with full client certificate
+ * Improved the response times on HTTP/0.9 (content without Content-length)
+ * Improved response granularity on above - using unbuffered BIO now
+ * Fixed problem with IE/SSL (SSL_set_shutdown)
+ * Avoid error messages on premature EOF from client
+ * Fixed HeadRemove code so all headers are checked without exception
+ * Improved autoconf detection
+ *
  * Revision 1.6  2003/11/30 22:56:26  roseg
  * Callback for RSA ephemeral keys:
  *     - generated in a separate thread
@@ -154,7 +218,8 @@ static char *rcs_id = "$Id: http.c,v 1.6 2003/11/30 22:56:26 roseg Rel $";
 /* HTTP error replies */
 static char *h500 = "500 Internal Server Error",
             *h501 = "501 Not Implemented",
-            *h503 = "503 Service Unavailable";
+            *h503 = "503 Service Unavailable",
+            *h414 = "414 Request URI too long";
 
 static char *err_head = "HTTP/1.0 %s\r\nContent-Type: text/html\r\nContent-Length: %d\r\n\r\n%s";
 static char *err_cont = "<html><head><title>%s</title></head><body><h1>%s</h1><p>%s</p></body></html>";
@@ -170,6 +235,7 @@ err_reply(BIO *c, char *head, char *txt)
     snprintf(cont, sizeof(cont), err_cont, head, head, txt);
     snprintf(rep, sizeof(rep), err_head, head, strlen(cont), cont);
     BIO_write(c, rep, strlen(rep));
+    BIO_flush(c);
     return;
 }
 
@@ -183,19 +249,15 @@ copy_bin(BIO *cl, BIO *be, long cont, long *res_bytes, int no_write)
     int         res;
 
     while(cont > 0L) {
-        if((res = BIO_read(cl, buf, cont > MAXBUF? MAXBUF: cont)) < 0) {
-            logmsg(LOG_WARNING, "copy_bin error reading: %s", strerror(errno));
+        if((res = BIO_read(cl, buf, cont > MAXBUF? MAXBUF: cont)) < 0)
             return -1;
-        }
-        if(res == 0) {
-            logmsg(LOG_WARNING, "copy_bin premature EOF: %s", strerror(errno));
+        else if(res == 0)
             return -2;
-        }
         if(!no_write)
-            if(BIO_write(be, buf, res) != res) {
-                logmsg(LOG_WARNING, "copy_bin error writing: %s", strerror(errno));
+            if(BIO_write(be, buf, res) != res)
                 return -3;
-            }
+        if(BIO_flush(be) != 1)
+            return -4;
         cont -= res;
         if(res_bytes)
             *res_bytes += res;
@@ -230,7 +292,8 @@ copy_chunks(BIO *cl, BIO *be, long *res_bytes, int no_write, long max_size)
 
     for(tot_size = 0L;;) {
         if(BIO_gets(cl, buf, MAXBUF) <= 0) {
-            logmsg(LOG_WARNING, "unexpected chunked EOF: %s", strerror(errno));
+            if(errno)
+                logmsg(LOG_WARNING, "unexpected chunked EOF: %s", strerror(errno));
             return -1;
         }
         strip_eol(buf);
@@ -255,14 +318,16 @@ copy_chunks(BIO *cl, BIO *be, long *res_bytes, int no_write, long max_size)
 
         if(cont > 0L) {
             if(copy_bin(cl, be, cont, res_bytes, no_write)) {
-                logmsg(LOG_WARNING, "error copy chunk cont: %s", strerror(errno));
+                if(errno)
+                    logmsg(LOG_WARNING, "error copy chunk cont: %s", strerror(errno));
                 return -4;
             }
         } else
             break;
         /* final CRLF */
         if(BIO_gets(cl, buf, MAXBUF) <= 0) {
-            logmsg(LOG_WARNING, "unexpected after chunk EOF: %s", strerror(errno));
+            if(errno)
+                logmsg(LOG_WARNING, "unexpected after chunk EOF: %s", strerror(errno));
             return -5;
         }
         strip_eol(buf);
@@ -277,14 +342,20 @@ copy_chunks(BIO *cl, BIO *be, long *res_bytes, int no_write, long max_size)
     /* possibly trailing headers */
     for(;;) {
         if(BIO_gets(cl, buf, MAXBUF) <= 0) {
-            logmsg(LOG_WARNING, "unexpected post-chunk EOF: %s", strerror(errno));
+            if(errno)
+                logmsg(LOG_WARNING, "unexpected post-chunk EOF: %s", strerror(errno));
             return -7;
         }
-        if(!no_write)
+        if(!no_write) {
             if(BIO_puts(be, buf) <= 0) {
                 logmsg(LOG_WARNING, "error post-chunk write: %s", strerror(errno));
                 return -8;
             }
+            if(BIO_flush(be) != 1) {
+                logmsg(LOG_WARNING, "copy_chunks flush error: %s", strerror(errno));
+                return -4;
+            }
+        }
         strip_eol(buf);
         if(!buf[0])
             break;
@@ -292,6 +363,8 @@ copy_chunks(BIO *cl, BIO *be, long *res_bytes, int no_write, long max_size)
     return 0;
 }
 
+static int  err_to = -1;
+
 /*
  * Time-out for client read/gets
  * the SSL manual says not to do it, but it works well enough anyway...
@@ -299,31 +372,51 @@ copy_chunks(BIO *cl, BIO *be, long *res_bytes, int no_write, long max_size)
 static long
 bio_callback(BIO *bio, int cmd, const char *argp, int argi, long argl, long ret)
 {
-    fd_set          socks;
-    struct timeval  to;
-    int             s, v;
+    struct pollfd   p;
+    int             to;
 
-    if(cmd == BIO_CB_READ)
-        for(;;) {
-            to.tv_sec = *((int *)BIO_get_callback_arg(bio));
-            to.tv_usec = 0;
-            BIO_get_fd(bio, &s);
-            FD_ZERO(&socks);
-            FD_SET(s, &socks);
-            if((v = select(s + 1, &socks, NULL, NULL, &to)) > 0)
-                break;
-            /* if select() returned 0 a time-out occured */
-            if(v == 0) {
-                /* logmsg(LOG_WARNING, "callback select timeout"); */
-                return -1;
+    if(cmd != BIO_CB_READ && cmd != BIO_CB_WRITE)
+        return ret;
+
+    /* a time-out already occured */
+    if((to = *((int *)BIO_get_callback_arg(bio)) * 1000) < 0) {
+        errno = ETIMEDOUT;
+        return -1;
+    }
+
+    for(;;) {
+        memset(&p, 0, sizeof(p));
+        BIO_get_fd(bio, &p.fd);
+        p.events = (cmd == BIO_CB_READ)? (POLLIN | POLLPRI): POLLOUT;
+        switch(poll(&p, 1, to)) {
+        case 1:
+            if(cmd == BIO_CB_READ) {
+                if((p.revents & POLLIN) || (p.revents & POLLPRI))
+                    /* there is readable data */
+                    return ret;
+                else
+                    errno = EIO;
+            } else {
+                if(p.revents & POLLOUT)
+                    /* data can be written */
+                    return ret;
+                else
+                    errno = ECONNRESET;
             }
-            /* if select() returned < 0 and it was not interrupted it's bad */
+            return -1;
+        case 0:
+            /* timeout - mark the BIO as unusable for the future */
+            BIO_set_callback_arg(bio, (char *)&err_to);
+            errno = ETIMEDOUT;
+            return 0;
+        default:
+            /* error */
             if(errno != EINTR) {
-                logmsg(LOG_WARNING, "callback select: %s", strerror(errno));
+                logmsg(LOG_WARNING, "callback poll: %s", strerror(errno));
                 return -2;
             }
         }
-    return ret;
+    }
 }
 
 /*
@@ -332,18 +425,14 @@ bio_callback(BIO *bio, int cmd, const char *argp, int argi, long argl, long ret)
 static int
 is_readable(BIO *bio, int to_wait)
 {
-    fd_set          socks;
-    struct timeval  to;
-    int             s;
+    struct pollfd   p;
 
     if(BIO_pending(bio) > 0)
         return 1;
-    s = BIO_get_fd(bio, NULL);
-    to.tv_sec = to_wait;
-    to.tv_usec = 0;
-    FD_ZERO(&socks);
-    FD_SET(s, &socks);
-    return(select(s + 1, &socks, NULL, NULL, &to) > 0);
+    memset(&p, 0, sizeof(p));
+    BIO_get_fd(bio, &p.fd);
+    p.events = POLLIN | POLLPRI;
+    return (poll(&p, 1, to_wait * 1000) > 0);
 }
 
 static void
@@ -359,7 +448,7 @@ free_headers(char **headers)
 }
 
 static char **
-get_headers(BIO *in)
+get_headers(BIO *in, BIO *cl)
 {
     char    **headers, buf[MAXBUF];
     int     res, n;
@@ -370,23 +459,36 @@ get_headers(BIO *in)
         if(buf[0])
             break;
     }
-    if(res <= 0)
+
+    if(res <= 0) {
+        /* this is expected to occur only on client reads */
+        /* logmsg(LOG_WARNING, "headers: bad starting read"); */
         return NULL;
+    } else if(res >= (MAXBUF - 1)) {
+        /* check for request length limit */
+        logmsg(LOG_WARNING, "headers: request URI too long");
+        err_reply(cl, h414, e414);
+        return NULL;
+    }
 
     if((headers = (char **)calloc(MAXHEADERS, sizeof(char *))) == NULL) {
         logmsg(LOG_WARNING, "headers: out of memory");
+        err_reply(cl, h500, e500);
         return NULL;
     }
 
     for(n = 0; n < MAXHEADERS; n++) {
-        if((headers[n] = strdup(buf)) == NULL) {
+        if((headers[n] = (char *)malloc(MAXBUF)) == NULL) {
             free_headers(headers);
             logmsg(LOG_WARNING, "header: out of memory");
+            err_reply(cl, h500, e500);
             return NULL;
         }
+        strncpy(headers[n], buf, MAXBUF - 1);
         if((res = BIO_gets(in, buf, MAXBUF)) <= 0) {
             free_headers(headers);
             logmsg(LOG_WARNING, "can't read header");
+            err_reply(cl, h500, e500);
             return NULL;
         }
         strip_eol(buf);
@@ -396,6 +498,7 @@ get_headers(BIO *in)
 
     free_headers(headers);
     logmsg(LOG_WARNING, "too many headers");
+    err_reply(cl, h500, e500);
     return NULL;
 }
 
@@ -482,8 +585,11 @@ URL_syntax(char *line)
 
 /* Cleanup code. This should really be in the pthread_cleanup_push, except for bugs in some implementations */
 #define clean_all() {   \
-    if(be != NULL) { BIO_flush(be); BIO_free_all(be); be = NULL; } \
-    if(cl != NULL) { BIO_flush(cl); BIO_free_all(cl); cl = NULL; } \
+    if(ssl != NULL) { BIO_ssl_shutdown(cl); BIO_ssl_shutdown(cl); BIO_ssl_shutdown(cl); } \
+    if(be != NULL) { BIO_flush(be); BIO_reset(be); BIO_free_all(be); be = NULL; } \
+    if(cl != NULL) { BIO_flush(cl); BIO_reset(cl); BIO_free_all(cl); cl = NULL; } \
+    if(x509 != NULL) { X509_free(x509); x509 = NULL; } \
+    if(ssl != NULL) { ERR_clear_error(); ERR_remove_state(0); } \
 }
 
 /*
@@ -492,16 +598,17 @@ URL_syntax(char *line)
 void *
 thr_http(void *arg)
 {
-    BIO                 *cl, *be, *bb;
+    BIO                 *cl, *be, *bb, *b64;
     X509                *x509;
-    SSL_CTX             *ctx;
     SSL                 *ssl;
     thr_arg             *a;
     struct in_addr      from_host;
     struct sockaddr_in  *srv, to_host;
-    int                 cl_11, be_11, res, chunked, n, sock, no_cont, skip, conn_closed, redir;
+    int                 cl_11, be_11, res, chunked, n, sock, no_cont, skip, conn_closed, redir,
+                        force_10;
     char                request[MAXBUF], response[MAXBUF], buf[MAXBUF], url[MAXBUF], loc_path[MAXBUF], **headers,
-                        headers_ok[MAXHEADERS], v_host[MAXBUF], referer[MAXBUF], u_agent[MAXBUF], *mh;
+                        headers_ok[MAXHEADERS], v_host[MAXBUF], referer[MAXBUF], u_agent[MAXBUF], u_name[MAXBUF],
+                        *mh;
     long                cont, res_bytes;
     time_t              req_start;
     regmatch_t          matches[4];
@@ -512,15 +619,24 @@ thr_http(void *arg)
     from_host = a->from_host;
     to_host = a->to_host;
     sock = a->sock;
-    ctx = a->ctx;
+    ssl = a->ssl;
     free(a);
 
-    n = 0;
+    n = 1;
     setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, (void *)&n, sizeof(n));
     l.l_onoff = 1;
-    l.l_linger = 30;
+    l.l_linger = 10;
     setsockopt(sock, SOL_SOCKET, SO_LINGER, (void *)&l, sizeof(l));
 
+#ifdef  TCP_LINGER2
+    n = 5;
+    setsockopt(sock, SOL_TCP, TCP_LINGER2, (void *)&n, sizeof(n));
+#endif
+
+    cl = NULL;
+    be = NULL;
+    x509 = NULL;
+
     if((cl = BIO_new_socket(sock, 1)) == NULL) {
         logmsg(LOG_WARNING, "BIO_new_socket failed");
         shutdown(sock, 2);
@@ -532,38 +648,38 @@ thr_http(void *arg)
         BIO_set_callback(cl, bio_callback);
     }
 
-    if(ctx != NULL) {
-        if((bb = BIO_new_ssl(ctx, 0)) == NULL) {
-            logmsg(LOG_WARNING, "BIO_new_ssl(ctx, 0) failed");
+    if(ssl != NULL) {
+        SSL_set_bio(ssl, cl, cl);
+        if((bb = BIO_new(BIO_f_ssl())) == NULL) {
+            logmsg(LOG_WARNING, "BIO_new(Bio_f_ssl()) failed");
+            BIO_reset(cl);
             BIO_free_all(cl);
             pthread_exit(NULL);
         }
+        BIO_set_ssl(bb, ssl, BIO_CLOSE);
         BIO_set_ssl_mode(bb, 0);
-        cl = BIO_push(bb, cl);
+        cl = bb;
         if(BIO_do_handshake(cl) <= 0) {
             /* no need to log every client without a certificate...
             logmsg(LOG_WARNING, "BIO_do_handshake with %s failed: %s", inet_ntoa(from_host),
                 ERR_error_string(ERR_get_error(), NULL));
             x509 = NULL;
             */
+            BIO_reset(cl);
             BIO_free_all(cl);
             pthread_exit(NULL);
         } else {
-            BIO_get_ssl(bb, &ssl);
-            if(ssl == NULL) {
-                logmsg(LOG_WARNING, "BIO_get_ssl failed");
-                BIO_free_all(cl);
-                pthread_exit(NULL);
-            }
-            if((x509 = SSL_get_peer_certificate(ssl)) != NULL && SSL_get_verify_result(ssl) != X509_V_OK) {
+            if((x509 = SSL_get_peer_certificate(ssl)) != NULL && https_headers < 3 && SSL_get_verify_result(ssl) != X509_V_OK) {
                 logmsg(LOG_WARNING, "Bad certificate from %s", inet_ntoa(from_host));
+                BIO_reset(cl);
                 BIO_free_all(cl);
                 pthread_exit(NULL);
             }
         }
         /*
          * This is dealt with in the handshake
-        if(https_headers > 1 && x509 == NULL) {
+        if(https_headers == 2 && x509 == NULL) {
+            BIO_reset(cl);
             BIO_free_all(cl);
             pthread_exit(NULL);
         }
@@ -574,25 +690,26 @@ thr_http(void *arg)
 
     if((bb = BIO_new(BIO_f_buffer())) == NULL) {
         logmsg(LOG_WARNING, "BIO_new(buffer) failed");
+        BIO_reset(cl);
         BIO_free_all(cl);
         pthread_exit(NULL);
     }
+    BIO_set_close(cl, BIO_CLOSE);
+    BIO_set_buffer_size(cl, MAXBUF);
     cl = BIO_push(bb, cl);
 
-    be = NULL;
-
     for(cl_11 = be_11 = 0;;) {
         req_start = time(NULL);
         res_bytes = 0L;
-        v_host[0] = referer[0] = u_agent[0] = '\0';
+        v_host[0] = referer[0] = u_agent[0] = u_name[0] = '\0';
         conn_closed = 0;
         for(n = 0; n < MAXHEADERS; n++)
             headers_ok[n] = 1;
-        if((headers = get_headers(cl)) == NULL) {
+        if((headers = get_headers(cl, cl)) == NULL) {
             if(!cl_11) {
-                logmsg(LOG_WARNING, "%d error read from %s: %s", pthread_self(), inet_ntoa(from_host),
-                    strerror(errno));
-                err_reply(cl, h500, e500);
+                if(errno)
+                    logmsg(LOG_WARNING, "error read from %s: %s", inet_ntoa(from_host), strerror(errno));
+                /* err_reply(cl, h500, e500); */
             }
             clean_all();
             pthread_exit(NULL);
@@ -629,7 +746,7 @@ thr_http(void *arg)
         }
 
         /* check other headers */
-        for(chunked = 0, cont = 0L, n = 1; n < MAXHEADERS && headers[n]; n++)
+        for(chunked = 0, cont = 0L, n = 1; n < MAXHEADERS && headers[n]; n++) {
             /* no overflow - see check_header for details */
             switch(check_header(headers[n], buf)) {
             case HEADER_HOST:
@@ -650,27 +767,63 @@ thr_http(void *arg)
                     conn_closed = 1;
                 break;
             case HEADER_TRANSFER_ENCODING:
-                if(!strcasecmp("chunked", buf))
-                    chunked = 1;
+                if(cont != 0L)
+                    headers_ok[n] = 0;
+                else if(!strcasecmp("chunked", buf))
+                    if(chunked)
+                        headers_ok[n] = 0;
+                    else
+                        chunked = 1;
                 break;
             case HEADER_CONTENT_LENGTH:
-                cont = atol(buf);
+                if(chunked)
+                    headers_ok[n] = 0;
+                else
+                    cont = atol(buf);
                 break;
             case HEADER_ILLEGAL:
                 if(log_level > 0)
                     logmsg(LOG_WARNING, "bad header from %s (%s)", inet_ntoa(from_host), headers[n]);
                 headers_ok[n] = 0;
                 break;
-            default:
-                /* maybe header to be removed */
-                if(head_off) {
-                    int i;
-
-                    for(i = 0; i < n_head_off; i++)
-                        headers_ok[n] = regexec(&head_off[i], headers[n], 0, NULL, 0);
-                }
-                break;
             }
+            if(headers_ok[n] && head_off) {
+                /* maybe header to be removed */
+                int i;
+
+                for(i = 0; i < n_head_off; i++)
+                    headers_ok[n] = regexec(&head_off[i], headers[n], 0, NULL, 0);
+            }
+            /* get User name */
+            if(!regexec(&AUTHORIZATION, headers[n], 2, matches, 0)) {
+                int inlen;
+
+                if((bb = BIO_new(BIO_s_mem())) == NULL) {
+                    logmsg(LOG_WARNING, "Can't alloc BIO_s_mem");
+                    continue;
+                }
+                if((b64 = BIO_new(BIO_f_base64())) == NULL) {
+                    logmsg(LOG_WARNING, "Can't alloc BIO_f_base64");
+                    BIO_free(bb);
+                    continue;
+                }
+                b64 = BIO_push(b64, bb);
+                BIO_write(bb, headers[n] + matches[1].rm_so, matches[1].rm_eo - matches[1].rm_so);
+                BIO_write(bb, "\n", 1);
+                if((inlen = BIO_read(b64, buf, MAXBUF - 1)) <= 0) {
+                    logmsg(LOG_WARNING, "Can't read BIO_f_base64");
+                    BIO_free(b64);
+                    continue;
+                }
+                BIO_free(b64);
+                if((mh = strchr(buf, ':')) == NULL) {
+                    logmsg(LOG_WARNING, "Unknown authentication");
+                    continue;
+                }
+                *mh = '\0';
+                strcpy(u_name, buf);
+            }
+        }
 
         /* possibly limited request size */
         if(max_req > 0L && cont > 0L && cont > max_req) {
@@ -684,12 +837,14 @@ thr_http(void *arg)
         if(be != NULL) {
             if(is_readable(be, 0)) {
                 /* The only way it's readable is if it's at EOF, so close it! */
+                BIO_reset(be);
                 BIO_free_all(be);
                 be = NULL;
             }
         }
         /* check that the requested URL still fits the old back-end */
         if(be != NULL && grp != get_grp(url, &headers[1])) {
+            BIO_reset(be);
             BIO_free_all(be);
             be = NULL;
         }
@@ -711,17 +866,17 @@ thr_http(void *arg)
                 clean_all();
                 pthread_exit(NULL);
             }
-            if(connect(sock, (struct sockaddr *)srv, (socklen_t)sizeof(*srv)) < 0) {
+            if(connect_nb(sock, (struct sockaddr *)srv, (socklen_t)sizeof(*srv)) < 0) {
                 logmsg(LOG_WARNING, "backend %s:%hd connect: %s",
                     inet_ntoa(srv->sin_addr), ntohs(srv->sin_port), strerror(errno));
                 close(sock);
                 kill_be(srv);
                 continue;
             }
-            n = 0;
+            n = 1;
             setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, (void *)&n, sizeof(n));
             l.l_onoff = 1;
-            l.l_linger = 30;
+            l.l_linger = 10;
             setsockopt(sock, SOL_SOCKET, SO_LINGER, (void *)&l, sizeof(l));
             if((be = BIO_new_socket(sock, 1)) == NULL) {
                 logmsg(LOG_WARNING, "BIO_new_socket server failed");
@@ -731,6 +886,7 @@ thr_http(void *arg)
                 clean_all();
                 pthread_exit(NULL);
             }
+            BIO_set_close(be, BIO_CLOSE);
             if(server_to > 0) {
                 BIO_set_callback_arg(be, (char *)&server_to);
                 BIO_set_callback(be, bio_callback);
@@ -741,6 +897,8 @@ thr_http(void *arg)
                 clean_all();
                 pthread_exit(NULL);
             }
+            BIO_set_buffer_size(bb, MAXBUF);
+            BIO_set_close(bb, BIO_CLOSE);
             be = BIO_push(bb, be);
         }
 
@@ -760,7 +918,7 @@ thr_http(void *arg)
         free_headers(headers);
 
         /* if SSL put additional headers for client certificate */
-        if(ctx != NULL) {
+        if(ssl != NULL) {
             SSL_CIPHER  *cipher;
 
             if(https_header != NULL)
@@ -772,7 +930,7 @@ thr_http(void *arg)
                     pthread_exit(NULL);
                 }
             if(https_headers > 0 && x509 != NULL && (bb = BIO_new(BIO_s_mem())) != NULL) {
-                X509_NAME_print(bb, X509_get_subject_name(x509), 16);
+                X509_NAME_print_ex(bb, X509_get_subject_name(x509), 8, XN_FLAG_ONELINE & ~ASN1_STRFLGS_ESC_MSB);
                 BIO_gets(bb, buf, MAXBUF);
                 if(BIO_printf(be, "X-SSL-Subject: %s\r\n", buf) <= 0) {
                     logmsg(LOG_WARNING, "error write X-SSL-Subject to %s:%hd: %s",
@@ -783,7 +941,7 @@ thr_http(void *arg)
                     pthread_exit(NULL);
                 }
 
-                X509_NAME_print(bb, X509_get_issuer_name(x509), 16);
+                X509_NAME_print_ex(bb, X509_get_issuer_name(x509), 8, XN_FLAG_ONELINE & ~ASN1_STRFLGS_ESC_MSB);
                 BIO_gets(bb, buf, MAXBUF);
                 if(BIO_printf(be, "X-SSL-Issuer: %s\r\n", buf) <= 0) {
                     logmsg(LOG_WARNING, "error write X-SSL-Issuer to %s:%hd: %s",
@@ -823,18 +981,40 @@ thr_http(void *arg)
                     clean_all();
                     pthread_exit(NULL);
                 }
-                BIO_free_all(bb);
-            }
-            if((cipher = SSL_get_current_cipher(ssl)) != NULL) {
-                SSL_CIPHER_description(cipher, buf, MAXBUF);
-                if(BIO_printf(be, "X-SSL-cipher: %s\r\n", buf) <= 0) {
-                    logmsg(LOG_WARNING, "error write X-SSL-cipher to %s:%hd: %s",
+                PEM_write_bio_X509(bb, x509);
+                BIO_gets(bb, buf, MAXBUF);
+                strip_eol(buf);
+                if(BIO_printf(be, "X-SSL-certificate: %s\r\n", buf) <= 0) {
+                    logmsg(LOG_WARNING, "error write X-SSL-certificate to %s:%hd: %s",
                         inet_ntoa(srv->sin_addr), ntohs(srv->sin_port), strerror(errno));
                     err_reply(cl, h500, e500);
                     BIO_free_all(bb);
                     clean_all();
                     pthread_exit(NULL);
                 }
+                while(BIO_gets(bb, buf, MAXBUF) > 0) {
+                    strip_eol(buf);
+                    if(BIO_printf(be, "\t%s\r\n", buf) <= 0) {
+                        logmsg(LOG_WARNING, "error write X-SSL-certificate to %s:%hd: %s",
+                            inet_ntoa(srv->sin_addr), ntohs(srv->sin_port), strerror(errno));
+                        err_reply(cl, h500, e500);
+                        BIO_free_all(bb);
+                        clean_all();
+                        pthread_exit(NULL);
+                    }
+                }
+                if((cipher = SSL_get_current_cipher(ssl)) != NULL) {
+                    SSL_CIPHER_description(cipher, buf, MAXBUF);
+                    strip_eol(buf);
+                    if(BIO_printf(be, "X-SSL-cipher: %s\r\n", buf) <= 0) {
+                        logmsg(LOG_WARNING, "error write X-SSL-cipher to %s:%hd: %s",
+                            inet_ntoa(srv->sin_addr), ntohs(srv->sin_port), strerror(errno));
+                        err_reply(cl, h500, e500);
+                        clean_all();
+                        pthread_exit(NULL);
+                    }
+                }
+                BIO_free_all(bb);
             }
         }
         /* put additional client IP header */
@@ -869,9 +1049,27 @@ thr_http(void *arg)
             pthread_exit(NULL);
         }
 
+        /*
+         * check on no_https_11:
+         *  - if 0 ignore
+         *  - if 1 and SSL force HTTP/1.0
+         *  - if 2 and SSL and MSIE force HTTP/1.0
+         */
+        switch(no_https_11) {
+        case 1:
+            force_10 = (ssl != NULL);
+            break;
+        case 2:
+            force_10 = (ssl != NULL && strstr(u_agent, "MSIE") != NULL);
+            break;
+        default:
+            force_10 = 0;
+            break;
+        }
+
         /* get the response */
         for(skip = 1; skip;) {
-            if((headers = get_headers(be)) == NULL) {
+            if((headers = get_headers(be, cl)) == NULL) {
                 logmsg(LOG_WARNING, "response error read from %s:%hd: %s",
                     inet_ntoa(srv->sin_addr), ntohs(srv->sin_port), strerror(errno));
                 err_reply(cl, h500, e500);
@@ -889,26 +1087,27 @@ thr_http(void *arg)
             /* check for redirection */
             redir = !regexec(&RESP_REDIR, response, 0, NULL, 0);
 
-            for(chunked = 0, cont = 0L, n = 1; n < MAXHEADERS && headers[n]; n++) {
+            for(chunked = 0, cont = -1L, n = 1; n < MAXHEADERS && headers[n]; n++) {
                 switch(check_header(headers[n], buf)) {
                 case HEADER_CONNECTION:
                     if(!strcasecmp("close", buf))
                         conn_closed = 1;
                     break;
                 case HEADER_TRANSFER_ENCODING:
-                    if(!strcasecmp("chunked", buf))
+                    if(!strcasecmp("chunked", buf)) {
                         chunked = 1;
+                        no_cont = 0;
+                    }
                     break;
                 case HEADER_CONTENT_LENGTH:
                     cont = atol(buf);
                     break;
                 case HEADER_LOCATION:
-                    if(redir && v_host[0] && is_be(buf, &to_host, loc_path)) {
+                    if(rewrite_redir && redir && v_host[0] && is_be(buf, &to_host, v_host, loc_path, grp)) {
                         snprintf(buf, MAXBUF, "Location: %s://%s/%s",
-                            (ctx == NULL? "http": "https"), v_host, loc_path);
+                            (ssl == NULL? "http": "https"), v_host, loc_path);
                         free(headers[n]);
                         if((headers[n] = strdup(buf)) == NULL) {
-                            headers[n] = NULL;
                             logmsg(LOG_WARNING, "rewrite Location - out of memory: %s", strerror(errno));
                             free_headers(headers);
                             clean_all();
@@ -926,7 +1125,8 @@ thr_http(void *arg)
             if(!skip)
                 for(n = 0; n < MAXHEADERS && headers[n]; n++) {
                     if(BIO_printf(cl, "%s\r\n", headers[n]) <= 0) {
-                        logmsg(LOG_WARNING, "error write to %s: %s", inet_ntoa(from_host), strerror(errno));
+                        if(errno)
+                            logmsg(LOG_WARNING, "error write to %s: %s", inet_ntoa(from_host), strerror(errno));
                         free_headers(headers);
                         clean_all();
                         pthread_exit(NULL);
@@ -937,6 +1137,12 @@ thr_http(void *arg)
             /* final CRLF */
             if(!skip)
                 BIO_puts(cl, "\r\n");
+            if(BIO_flush(cl) != 1) {
+                if(errno)
+                    logmsg(LOG_WARNING, "error flush headers to %s: %s", inet_ntoa(from_host), strerror(errno));
+                clean_all();
+                pthread_exit(NULL);
+            }
 
             if(!no_cont) {
                 /* ignore this if request was HEAD or similar */
@@ -947,37 +1153,75 @@ thr_http(void *arg)
                         clean_all();
                         pthread_exit(NULL);
                     }
-                } else if(cont > 0L) {
-                    /* had Content-length, so do raw reads/writes for the length */
+                } else if(cont >= 0L) {
+                    /* may have had Content-length, so do raw reads/writes for the length */
                     if(copy_bin(be, cl, cont, &res_bytes, skip)) {
-                        logmsg(LOG_WARNING, "error copy server cont: %s", strerror(errno));
+                        if(errno)
+                            logmsg(LOG_WARNING, "error copy server cont: %s", strerror(errno));
                         clean_all();
                         pthread_exit(NULL);
                     }
                 } else if(!skip) {
-                    /*
-                     * old-style response - content until EOF
-                     * also implies the client may not use HTTP/1.1
-                     */
-                    cl_11 = be_11 = 0;
+                    if(is_readable(be, SERVER_TO)) {
+                        char    one;
+                        BIO     *be_unbuf;
+                        /*
+                         * old-style response - content until EOF
+                         * also implies the client may not use HTTP/1.1
+                         */
+                        cl_11 = be_11 = 0;
 
-                    while((res = BIO_read(be, buf, MAXBUF)) > 0) {
-                        if(BIO_write(cl, buf, res) != res) {
-                            logmsg(LOG_WARNING, "error copy response body: %s", strerror(errno));
+                        /*
+                         * first read whatever is already in the input buffer
+                         */
+                        while(BIO_pending(be)) {
+                            if(BIO_read(be, &one, 1) != 1) {
+                                logmsg(LOG_WARNING, "error read response pending: %s", strerror(errno));
+                                clean_all();
+                                pthread_exit(NULL);
+                            }
+                            if(BIO_write(cl, &one, 1) != 1) {
+                                if(errno)
+                                    logmsg(LOG_WARNING, "error write response pending: %s", strerror(errno));
+                                clean_all();
+                                pthread_exit(NULL);
+                            }
+                            res_bytes++;
+                        }
+                        BIO_flush(cl);
+
+                        /*
+                         * find the socket BIO in the chain
+                         */
+                        if((be_unbuf = BIO_find_type(be, BIO_TYPE_SOCKET)) == NULL) {
+                            logmsg(LOG_WARNING, "error get unbuffered: %s", strerror(errno));
                             clean_all();
                             pthread_exit(NULL);
-                        } else
-                            res_bytes += res;
+                        }
+
+                        /*
+                         * copy till EOF
+                         */
+                        while((res = BIO_read(be_unbuf, buf, MAXBUF)) > 0) {
+                            if(BIO_write(cl, buf, res) != res) {
+                                if(errno)
+                                    logmsg(LOG_WARNING, "error copy response body: %s", strerror(errno));
+                                clean_all();
+                                pthread_exit(NULL);
+                            } else {
+                                res_bytes += res;
+                                BIO_flush(cl);
+                            }
+                        }
                     }
                 }
-            }
-
-            if(!skip)
                 if(BIO_flush(cl) != 1) {
-                    logmsg(LOG_WARNING, "error flush to %s: %s", inet_ntoa(from_host), strerror(errno));
+                    if(errno)
+                        logmsg(LOG_WARNING, "error final flush to %s: %s", inet_ntoa(from_host), strerror(errno));
                     clean_all();
                     pthread_exit(NULL);
                 }
+            }
         }
 
         /* log what happened */
@@ -995,22 +1239,23 @@ thr_http(void *arg)
             break;
         case 3:
             if(v_host[0])
-                logmsg(LOG_INFO, "%s %s - - [%s] \"%s\" %c%c%c %s \"%s\" \"%s\"", v_host, inet_ntoa(from_host),
-                    log_time(req_start), request, response[9], response[10], response[11], log_bytes(res_bytes),
-                    referer, u_agent);
+                logmsg(LOG_INFO, "%s %s - %s [%s] \"%s\" %c%c%c %s \"%s\" \"%s\"", v_host, inet_ntoa(from_host),
+                    u_name[0]? u_name: "-", log_time(req_start), request, response[9], response[10], response[11],
+                    log_bytes(res_bytes), referer, u_agent);
             else
-                logmsg(LOG_INFO, "%s - - [%s] \"%s\" %c%c%c %s \"%s\" \"%s\"", inet_ntoa(from_host),
-                    log_time(req_start), request, response[9], response[10], response[11], log_bytes(res_bytes),
-                    referer, u_agent);
+                logmsg(LOG_INFO, "%s - %s [%s] \"%s\" %c%c%c %s \"%s\" \"%s\"", inet_ntoa(from_host),
+                    u_name[0]? u_name: "-", log_time(req_start), request, response[9], response[10], response[11],
+                    log_bytes(res_bytes), referer, u_agent);
             break;
         case 4:
-            logmsg(LOG_INFO, "%s - - [%s] \"%s\" %c%c%c %s \"%s\" \"%s\"", inet_ntoa(from_host),
-                log_time(req_start), request, response[9], response[10], response[11], log_bytes(res_bytes),
-                referer, u_agent);
+            logmsg(LOG_INFO, "%s - %s [%s] \"%s\" %c%c%c %s \"%s\" \"%s\"", inet_ntoa(from_host),
+                log_time(req_start), u_name[0]? u_name: "-", request, response[9], response[10], response[11],
+                log_bytes(res_bytes), referer, u_agent);
             break;
         }
 
         if(!be_11) {
+            BIO_reset(be);
             BIO_free_all(be);
             be = NULL;
         }
@@ -1022,10 +1267,16 @@ thr_http(void *arg)
          *      or
          *  - this is an SSL connection and we had a NoHTTPS11 directive
          */
-        if(!cl_11 || conn_closed || (ctx != NULL && no_https_11))
+        if(!cl_11 || conn_closed || force_10)
             break;
     }
 
+    /*
+     * This may help with some versions of IE with a broken channel shutdown
+     */
+    if(ssl != NULL)
+        SSL_set_shutdown(ssl, SSL_SENT_SHUTDOWN | SSL_RECEIVED_SHUTDOWN);
+
     clean_all();
     pthread_exit(NULL);
 }
