@@ -26,10 +26,24 @@
  * EMail: roseg@apsis.ch
  */
 
-static char *rcs_id = "$Id: pound.c,v 1.9 2005/06/01 15:01:53 roseg Rel roseg $";
+static char *rcs_id = "$Id: pound.c,v 1.10 2006/02/01 11:19:53 roseg Rel $";
 
 /*
  * $Log: pound.c,v $
+ * Revision 1.10  2006/02/01 11:19:53  roseg
+ * Enhancements:
+ *   added NoDaemon configuration directive (replaces compile-time switch)
+ *   added LogFacility configuration directive (replaces compile-time switch)
+ *   added user name logging
+ *
+ * Bug fixes:
+ *   fixed problem with the poll() code
+ *   fixed problem with empty list in gethostbyname()
+ *   added call to setsid() if daemon
+ *   conflicting headers are removed (Content-length - Transfer-encoding)
+ *
+ * Last release in the 1.x series.
+ *
  * Revision 1.9  2005/06/01 15:01:53  roseg
  * Enhancements:
  *   Added the VerifyList configuration flag (CA root certs + CRL)
