@@ -26,10 +26,74 @@
  * EMail: roseg@apsis.ch
  */
 
-static char *rcs_id = "$Id: svc.c,v 1.6 2003/11/30 22:56:26 roseg Rel $";
+static char *rcs_id = "$Id: svc.c,v 1.10 2006/02/01 11:19:54 roseg Rel $";
 
 /*
  * $Log: svc.c,v $
+ * Revision 1.10  2006/02/01 11:19:54  roseg
+ * Enhancements:
+ *   added NoDaemon configuration directive (replaces compile-time switch)
+ *   added LogFacility configuration directive (replaces compile-time switch)
+ *   added user name logging
+ *
+ * Bug fixes:
+ *   fixed problem with the poll() code
+ *   fixed problem with empty list in gethostbyname()
+ *   added call to setsid() if daemon
+ *   conflicting headers are removed (Content-length - Transfer-encoding)
+ *
+ * Last release in the 1.x series.
+ *
+ * Revision 1.9  2005/06/01 15:01:54  roseg
+ * Enhancements:
+ *   Added the VerifyList configuration flag (CA root certs + CRL)
+ *   CRL checking code
+ *   RewriteRedirect 2 - ignores port value for host matching
+ *   Added -c flag (check-only mode)
+ *   Added -v flag (verbose mode)
+ *   Added -p flag for pid file name
+ *
+ * Bug fixes:
+ *   fixed a potential buffer overflow problem (in checking the Host header)
+ *   added call to SSL_library_init
+ *   added a check for MSIE before forcing SSL shutdown
+ *   X-SSL-Cipher header is added only if HTTPSHeaders is non-zero
+ *   added code for shorter linger on badly closed connections (IE work-around)
+ *   fixed the locking for session checking (mutex_lock/unlock)
+ *
+ * Revision 1.8  2004/11/04 13:37:07  roseg
+ * Changes:
+ * - added support for non-blocking connect(2)
+ * - added support for 414 - Request URI too long
+ * - added RedirectRewrite directive - to prevent redirect changes
+ * - added support for NoHTTPS11 value 2 (for MSIE clients only)
+ * - added support for HTTPSHeaders 3 (no verify)
+ *
+ * Problems fixed:
+ * - fixed bug if multiple listening ports/addresses
+ * - fixed memory leak in SSL
+ * - flush stdout (if used) after each log message
+ * - assumes only 304, 305 and 306 codes to have no content
+ * - fixed problem with delays in 302 without content
+ * - fixed problem with time-outs in HTTPS
+ *
+ * Enhancements:
+ * - improved threads detection code in autoconf
+ * - added supervisor process disable configuration flag
+ * - tweak for the Location rewriting code (only look at current GROUP)
+ * - improved print-out for client certificate information
+ *
+ * Revision 1.7  2004/03/24 06:59:59  roseg
+ * Fixed bug in X-SSL-CIPHER description
+ * Changed README to stx format for consistency
+ * Addedd X-SSL-certificate with full client certificate
+ * Improved the response times on HTTP/0.9 (content without Content-length)
+ * Improved response granularity on above - using unbuffered BIO now
+ * Fixed problem with IE/SSL (SSL_set_shutdown)
+ * Avoid error messages on premature EOF from client
+ * Fixed HeadRemove code so all headers are checked without exception
+ * Improved autoconf detection
+ *
  * Revision 1.6  2003/11/30 22:56:26  roseg
  * Callback for RSA ephemeral keys:
  *     - generated in a separate thread
@@ -152,9 +216,10 @@ logmsg(int priority, char *fmt, ...)
     vsnprintf(buf, MAXBUF, fmt, ap);
     va_end(ap);
 #ifdef  NO_SYSLOG
-    if(priority == LOG_INFO)
+    if(priority == LOG_INFO) {
         printf("%s\n", buf);
-    else {
+        fflush(stdout);
+    } else {
         char    t_stamp[32];
         time_t  now;
 
@@ -163,7 +228,10 @@ logmsg(int priority, char *fmt, ...)
         fprintf(stderr, "%s: %s\n", t_stamp, buf);
     }
 #else
-    syslog(priority, "%s", buf);
+    if(print_log)
+        printf("%s\n", buf);
+    else
+        syslog(log_facility | priority, "%s", buf);
 #endif
     return;
 }
@@ -180,9 +248,10 @@ va_dcl
     vsnprintf(buf, MAXBUF, fmt, ap);
     va_end(ap);
 #ifdef  NO_SYSLOG
-    if(priority == LOG_INFO)
+    if(priority == LOG_INFO) {
         printf("%s\n", buf);
-    else {
+        fflush(stdout);
+    } else {
         char    t_stamp[32];
         time_t  now;
 
@@ -191,7 +260,10 @@ va_dcl
         fprintf(stderr, "%s: %s\n", t_stamp, buf);
     }
 #else
-    syslog(priority, "%s", buf);
+    if(print_log)
+        printf("%s\n", buf);
+    else
+        syslog(log_facility | priority, "%s", buf);
 #endif
     return;
 }
@@ -485,10 +557,10 @@ get_be(GROUP *g, struct in_addr from_host, char *url, char **headers)
     if(g->tot_pri == 0)
         return NULL;
 
+    pthread_mutex_lock(&g->mut);
     key = get_key(g, from_host, url, headers);
     if(key != NULL && g->sess_to > 0) {
         /* check for session, add it if necessary */
-        pthread_mutex_lock(&g->mut);
         if((sp = sess_find(g->sessions, key)) == NULL) {
             /* no session yet - create one */
             orig = n = rand() % g->tot_pri;
@@ -510,7 +582,6 @@ get_be(GROUP *g, struct in_addr from_host, char *url, char **headers)
             } else
                 res = NULL;
         }
-        pthread_mutex_unlock(&g->mut);
     } else {
         if(g->sess_to < 0) {
             /* "sticky" mappings */
@@ -537,6 +608,7 @@ get_be(GROUP *g, struct in_addr from_host, char *url, char **headers)
         else
             res = NULL;
     }
+    pthread_mutex_unlock(&g->mut);
 
     return res;
 }
@@ -551,11 +623,13 @@ upd_session(GROUP *g, char **headers, struct sockaddr_in  *srv)
     char            *key;
     int             n;
 
-    memset(&dummy, 0, sizeof(dummy));
-    if(g->sess_type != SessCOOKIE || (key = get_key(g, dummy, "", headers)) == NULL)
-        return;
-    /* probably found a Set-cookie, so we may have to create a session here */
     pthread_mutex_lock(&g->mut);
+    memset(&dummy, 0, sizeof(dummy));
+    if(g->sess_type != SessCOOKIE || (key = get_key(g, dummy, "", headers)) == NULL) {
+        pthread_mutex_unlock(&g->mut);
+        return;
+    }
+    /* probably found a Set-cookie, so we may have to create a session here */
     if(sess_find(g->sessions, key) == NULL) {
         /* no session yet - create one */
         for(n = 0; n < g->tot_pri; n++)
@@ -597,32 +671,18 @@ kill_be(struct sockaddr_in *be)
     return;
 }
 
-/*
- * compare two host addresses - only address and port matter
- * this might be a bug in the Linux TCP stack!
- */
-static int
-addrcmp(struct sockaddr_in *a1, struct sockaddr_in *a2)
-{
-    int res;
-
-    if((res = memcmp(&a1->sin_addr, &a2->sin_addr, sizeof(a1->sin_addr))))
-        return res;
-    return memcmp(&a1->sin_port, &a2->sin_port, sizeof(a1->sin_port));
-}
-
 /*
  * Find if a host is in our list of back-ends
  */
 int
-is_be(char *location, struct sockaddr_in *to_host, char *path)
+is_be(char *location, struct sockaddr_in *to_host, char *v_hostport, char *path, GROUP *grp)
 {
     int     i, n;
     GROUP   *g;
     struct sockaddr_in  addr;
     struct hostent      *he;
     regmatch_t          matches[4];
-    char                *proto, *host, *port;
+    char                *proto, *host, *port, *v_host, *v_port;
 
     /* split the location into its fields */
     if(regexec(&LOCATION, location, 4, matches, 0))
@@ -634,25 +694,39 @@ is_be(char *location, struct sockaddr_in *to_host, char *path)
     if((port = strchr(host, ':')) != NULL)
         *port++ = '\0';
 
+    /*
+     * rewrite if hostname in Host: and Location: are the same
+     * applies only if RewriteRedir is 2
+     */
+    if(rewrite_redir == 2 && (v_host = strdup(v_hostport)) != NULL) {
+        if ((v_port = strchr(v_host, ':')) != NULL)
+            *v_port++ = '\0';
+        if(strcmp(host, v_host) == 0) {
+            free(v_host);
+            return 1;
+        }
+        free(v_host);
+    }
+
     memset(&addr, 0, sizeof(addr));
     addr.sin_family = AF_INET;
-    if((he = gethostbyname(host)) == NULL)
+    if((he = gethostbyname(host)) == NULL || he->h_addr_list[0] == NULL)
         return 0;
     memcpy(&addr.sin_addr.s_addr, he->h_addr_list[0], sizeof(addr.sin_addr.s_addr));
     if(port != NULL)
         addr.sin_port = (in_port_t)htons(atoi(port));
-    else if(strcmp(proto, "http") == 0)
-        addr.sin_port = (in_port_t)htons(80);
-    else /* if(strcmp(proto, "https") == 0) */
+    else if(strncmp(proto, "https", 5) == 0)
         addr.sin_port = (in_port_t)htons(443);
+    else
+        addr.sin_port = (in_port_t)htons(80);
 
-    if(addrcmp(to_host, &addr) == 0)
+    if(memcmp(&to_host->sin_addr, &addr.sin_addr, sizeof(addr.sin_addr)) == 0
+    && memcmp(&to_host->sin_port, &addr.sin_port, sizeof(addr.sin_port)) == 0)
         return 1;
-    for(n = 0; (g = groups[n]) != NULL; n++) {
-        for(i = 0; i < g->tot_pri; i++)
-            if(addrcmp(&(g->backend_addr[i].addr), &addr) == 0)
-                return 1;
-    }
+    for(i = 0; i < grp->tot_pri; i++)
+        if(memcmp(&grp->backend_addr[i].addr.sin_addr, &addr.sin_addr, sizeof(addr.sin_addr)) == 0
+        && memcmp(&grp->backend_addr[i].addr.sin_port, &addr.sin_port, sizeof(addr.sin_port)) == 0)
+            return 1;
     return 0;
 }
 
@@ -667,10 +741,78 @@ add_port(char *host, struct sockaddr_in *to_host)
     if(strchr(host, ':') != NULL)
         /* the host already contains a port */
         return NULL;
-    sprintf(res, "Host: %s:%hd", host, ntohs(to_host->sin_port));
+    snprintf(res, MAXBUF - 1, "Host: %s:%hd", host, ntohs(to_host->sin_port));
     return strdup(res);
 }
 
+/*
+ * Non-blocking connect(). Does the same as connect(2) but ensures
+ * it will time-out after a much shorter time period SERVER_TO
+ */
+int
+connect_nb(int sockfd, struct sockaddr *serv_addr, socklen_t addrlen)
+{
+    int             flags, res, error;
+    socklen_t       len;
+    struct pollfd   p;
+
+    if((flags = fcntl(sockfd, F_GETFL, 0)) < 0) {
+        logmsg(LOG_ERR, "fcntl GETFL failed: %s", strerror(errno));
+        return -1;
+    }
+    if(fcntl(sockfd, F_SETFL, flags | O_NONBLOCK) < 0) {
+        logmsg(LOG_ERR, "fcntl SETFL failed: %s", strerror(errno));
+        return -1;
+    }
+
+    error = 0;
+    if((res = connect(sockfd, serv_addr, addrlen)) < 0)
+        if(errno != EINPROGRESS)
+            return (-1);
+
+    if(res == 0) {
+        /* connect completed immediately (usually localhost) */
+        if(fcntl(sockfd, F_SETFL, flags) < 0) {
+            logmsg(LOG_ERR, "fcntl reSETFL failed: %s", strerror(errno));
+            return -1;
+        }
+        return 0;
+    }
+
+    memset(&p, 0, sizeof(p));
+    p.fd = sockfd;
+    p.events = POLLOUT;
+    if((res = poll(&p, 1, SERVER_TO * 1000)) != 1) {
+        if(res == 0) {
+            /* timeout */
+            errno = ETIMEDOUT;
+        }
+        return -1;
+    }
+
+    /* socket is writeable == operation completed */
+    len = sizeof(error);
+    if(getsockopt(sockfd, SOL_SOCKET, SO_ERROR, &error, &len) < 0) {
+        logmsg(LOG_ERR, "getsockopt failed: %s", strerror(errno));
+        return -1;
+    }
+
+    /* restore file status flags */
+    if(fcntl(sockfd, F_SETFL, flags) < 0) {
+        logmsg(LOG_ERR, "fcntl reSETFL failed: %s", strerror(errno));
+        return -1;
+    }
+
+    if(error) {
+        /* getsockopt() shows an error */
+        errno = error;
+        return -1;
+    }
+
+    /* really connected */
+    return 0;
+}
+
 /*
  * Prune the expired sessions and dead hosts from the table;
  * runs every session_to seconds (if needed)
@@ -726,7 +868,7 @@ thr_resurect(void *arg)
                 if((sock = socket(PF_INET, SOCK_STREAM, 0)) < 0)
                     continue;
                 addr = g->backend_addr[i].alive_addr;
-                if(connect(sock, (struct sockaddr *)&addr, (socklen_t)sizeof(addr)) != 0) {
+                if(connect_nb(sock, (struct sockaddr *)&addr, (socklen_t)sizeof(addr)) != 0) {
                     kill_be(&g->backend_addr[i].addr);
                     logmsg(LOG_ERR,"BackEnd %s is dead", inet_ntoa(g->backend_addr[i].addr.sin_addr));
                 }
@@ -745,7 +887,7 @@ thr_resurect(void *arg)
                     addr = g->backend_addr[i].addr;
                 else
                     addr = g->backend_addr[i].alive_addr;
-                if(connect(sock, (struct sockaddr *)&addr, (socklen_t)sizeof(addr)) == 0) {
+                if(connect_nb(sock, (struct sockaddr *)&addr, (socklen_t)sizeof(addr)) == 0) {
                     pthread_mutex_lock(&g->mut);
                     addr = g->backend_addr[i].addr;
                     for(j = i; j < g->tot_pri; j++)
@@ -780,6 +922,27 @@ RSA_tmp_callback(SSL *ssl, int is_export, int keylength)
     return res;
 }
 
+/*
+ * Pre-generate ephemeral RSA keys
+ */
+init_RSAgen()
+{
+    int n;
+
+    for(n = 0; n < N_RSA_KEYS; n++) {
+        if((RSA512_keys[n] = RSA_generate_key(512, RSA_F4, NULL, NULL)) == NULL) {
+            logmsg(LOG_ERR,"RSA_generate(%d, 512) failed", n);
+            return -1;
+        }
+        if((RSA1024_keys[n] = RSA_generate_key(1024, RSA_F4, NULL, NULL)) == NULL) {
+            logmsg(LOG_ERR,"RSA_generate(%d, 1024) failed", n);
+            return -2;
+        }
+    }
+    pthread_mutex_init(&RSA_mut, NULL);
+    return 0;
+}
+
 /*
  * Periodically regenerate ephemeral RSA keys
  * runs every T_RSA_KEYS seconds
@@ -789,15 +952,6 @@ thr_RSAgen(void *arg)
 {
     int n;
 
-    pthread_mutex_init(&RSA_mut, NULL);
-    pthread_mutex_lock(&RSA_mut);
-fprintf(stderr, "start RSA gen\n");
-    for(n = 0; n < N_RSA_KEYS; n++) {
-        RSA512_keys[n] = RSA_generate_key(512, RSA_F4, NULL, NULL);
-        RSA1024_keys[n] = RSA_generate_key(1024, RSA_F4, NULL, NULL);
-    }
-fprintf(stderr, "done RSA gen\n");
-    pthread_mutex_unlock(&RSA_mut);
     for(;;) {
         sleep(T_RSA_KEYS);
         pthread_mutex_lock(&RSA_mut);
